/*
 * buffer.c
 *
 * routines of input
 * Copyright (c) 1988, 89, 90, 91, 92, 93 Miguel Santana
 * Copyright (c) 1995, 96, 97, 98 Akim Demaille, Miguel Santana
 * $Id: buffer.c,v 1.72 1998/03/04 16:20:56 demaille Exp $
 */

/*
 * This file is part of a2ps.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * $Id: buffer.c,v 1.72 1998/03/04 16:20:56 demaille Exp $
 */

#include "a2ps.h"
#include "buffer.h"
#include "routines.h"
#include "xobstack.h"
#include "options.h"

/****************************************************************/
/*		 	Handling of the various eol styles	*/
/****************************************************************/
const char *
eol_to_string (enum eol_e eol)
{
  switch (eol) {
  case eol_r:
    return "\\r";
    
  case eol_n:
    return "\\n";
    
  case eol_rn:
    return "\\r\\n";
    
  case eol_nr:
    return "\\n\\r";
    
  case eol_auto:
    /* TRANS: the type of the end-of-line rules, is `any type', i.e.
       lines ended by \n, or \r, or \r\n, or \n\r are OK */
    return _("any type");

  default:
    error (1, 0, "bad switch in eol_to_string");
  }
  return NULL;	/* -Wall */
}

/*
 * What about the booleans
 */
static const char *const eol_args[] =
{
  "r", "mac",
  "n", "unix",
  "nr",
  "rn", "pc",
  "auto", "any", "4x4",
  0
};

static const enum eol_e eol_types[] =
{
  eol_r, eol_r, 
  eol_n, eol_n,
  eol_nr,
  eol_rn, eol_nr, 
  eol_auto, eol_auto, eol_auto
};


enum eol_e
option_string_to_eol (const char * option,
		      const char * arg)
{
  int i = a2ps_get_symbolic_value (eol_args, eol_types,
				   option, arg);
  return eol_types[i];
}

/****************************************************************/
/*		 	struct buffer Service routines 		*/
/****************************************************************/
static inline struct buffer *
buffer_internal_new (FILE * stream, const uchar * string,
		     enum eol_e eol)
{
  struct buffer * res = ALLOC (struct buffer, 1);
  res->stream = stream;
  res->stream_string = string;
  res->offset = 0;
  res->content = NULL;
  res->eol = eol;
  res->lower_case = FALSE;	/* By default, not lower case version	*/
  res->value = NULL;
  res->line = 0;
  res->allocsize = 0;
  res->len = 0;
  res->curr = 0;
  res->input_stack = ALLOC (struct obstack, 1);
  obstack_init (res->input_stack);
  return res;
}

struct buffer *
buffer_stream_new (FILE * stream, enum eol_e eol)
{
  return buffer_internal_new (stream, NULL, eol);
}

struct buffer *
buffer_string_new (const uchar * string, enum eol_e eol)
{
  return buffer_internal_new (NULL, string, eol);
}

void
buffer_self_print (struct buffer * buffer, FILE * stream)
{
  if (buffer->stream) {
    fprintf (stream, "A stream buffer.\n");
  } else {
    fprintf (stream, "A string buffer.  Offset %u\n", buffer->offset);
  }
  fprintf (stream, "Len = %d, Lower case = %d, Line = %d\n",
	   buffer->len, buffer->lower_case, buffer->line);
  if (buffer->len)
    fprintf (stream, "Content = `%s'\n", buffer->content);
}

void
buffer_free (struct buffer * buffer)
{
  /* VALUE is malloc'd only if BUFFER->LOWER_CASE */
  if (buffer->lower_case)
    FREE (buffer->value);
  /* I don't know how this one should be used */
/*  obstack_free (buffer->input_stack, NULL);*/
  FREE (buffer->input_stack);
  FREE (buffer);
}

void
buffer_set_lower_case (struct buffer * buffer, int sensitive)
{
  buffer->lower_case = sensitive;
}

/*
 * Get a line from BUFFER->STREAM.
 */
static inline void
buffer_stream_get_line (struct buffer * buffer)
{
  register int c, d;

  for (;;) {
    c = getc (buffer->stream);
    
      switch (c) {
      case '\n':
	switch (buffer->eol) {
	case eol_r:
	case eol_rn:
	  /* \n plain char */
	  goto stream_plain_char;

	case eol_auto:
	  /* If the next char is a \r, eat it */
	  if ((d = getc (buffer->stream)) != '\r')
	    ungetc (d, buffer->stream);
	  break;

	case eol_n:
	  /* This is a good eol */
	  break;
	  
	case eol_nr:
	  if ((d = getc (buffer->stream)) != '\r') {
	    /* This is \n, but eol is \n\r: make it a plain char */
	    ungetc (d, buffer->stream);
	    goto stream_plain_char;
	  }
	  /* This is eol=\n\r, just return \n */
	  break;
	}
	/* Nul terminate and end */
	obstack_1grow (buffer->input_stack, c);
	buffer->len = obstack_object_size (buffer->input_stack);
	obstack_1grow (buffer->input_stack, '\0');
	buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	return;	   

      case '\r':
	switch (buffer->eol) {
	case eol_n:
	case eol_nr:
	  /* \r plain char */
	  goto stream_plain_char;

	case eol_r:
	  /* This is a good eol, but the lib uses \n */
	  c = '\n';
	  break;
	  
	case eol_auto:
	  /* If the next char is a \n, eat it */
	  if ((d = getc (buffer->stream)) != '\n')
	    ungetc (d, buffer->stream);
	  c = '\n';
	  break;

	case eol_rn:
	  if ((d = getc (buffer->stream)) != '\n') {
	    /* This is \r, but eol is \r\n: make it a plain char */
	    ungetc (d, buffer->stream);
	    goto stream_plain_char;
	  }
	  /* This is eol = \r\n: just return \n */
	  c = '\n';
	  break;
	}
	/* Nul terminate and end */
	obstack_1grow (buffer->input_stack, c);
	buffer->len = obstack_object_size (buffer->input_stack);
	obstack_1grow (buffer->input_stack, '\0');
	buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	return;	   

      case EOF:
	/* Nul terminate and end */
	if ((buffer->len = obstack_object_size (buffer->input_stack))) {
	  /* The last line is not eol terminated */
	  obstack_1grow (buffer->input_stack, '\0');
	  buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	}
	return;
	
      default:
      stream_plain_char:
	obstack_1grow (buffer->input_stack, c);
	break;
      }
  }
}

/*
 * Get a line from BUFFER->STREAM_STRING
 *
 * Note that we could have made it destuctive.
 * But would have caused problem if some day we want to use
 * mmap.
 */
static inline void
buffer_string_get_line (struct buffer * buffer)
{
  register int c, d;

  for (;;) {
    c = buffer->stream_string [buffer->offset++];
    
      switch (c) {
      case '\n':
	switch (buffer->eol) {
	case eol_r:
	case eol_rn:
	  /* \n plain char */
	  goto string_plain_char;

	case eol_auto:
	  /* If the next char is a \r, eat it */
	  if ((d = buffer->stream_string [buffer->offset++]) != '\r')
	    buffer->offset--;
	  break;

	case eol_n:
	  /* This is a good eol */
	  break;
	  
	case eol_nr:
	  if ((d = buffer->stream_string [buffer->offset++]) != '\r') {
	    /* This is \n, but eol is \n\r: make it a plain char */
	    buffer->offset--;
	    goto string_plain_char;
	  }
	  /* This is eol=\n\r, just return \n */
	  break;
	}
	/* Nul terminate and end */
	obstack_1grow (buffer->input_stack, c);
	buffer->len = obstack_object_size (buffer->input_stack);
	obstack_1grow (buffer->input_stack, '\0');
	buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	return;	   
	
      case '\r':
	switch (buffer->eol) {
	case eol_n:
	case eol_nr:
	  /* \r plain char */
	  goto string_plain_char;

	case eol_r:
	  /* This is a good eol, but the lib uses \n */
	  c = '\n';
	  break;
	  
	case eol_auto:
	  /* If the next char is a \n, eat it */
	  if ((d = buffer->stream_string [buffer->offset++]) != '\n')
	    buffer->offset--;
	  c = '\n';
	  break;

	case eol_rn:
	  if ((d = buffer->stream_string [buffer->offset++]) != '\n') {
	    /* This is \r, but eol is \r\n: make it a plain char */
	    buffer->offset--;
	    goto string_plain_char;
	  }
	  /* This is eol = \r\n: just return \n */
	  c = '\n';
	  break;
	}
	/* Nul terminate and end */
	obstack_1grow (buffer->input_stack, c);
	buffer->len = obstack_object_size (buffer->input_stack);
	obstack_1grow (buffer->input_stack, '\0');
	buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	return;	   

      case '\0':
	/* Nul terminate and end */
	if ((buffer->len = obstack_object_size (buffer->input_stack))) {
	  /* Make sure the next time we will fall onto the \0
	   * which is the eol char */
	  buffer->offset--;
	  /* The last line is not eol terminated */
	  obstack_1grow (buffer->input_stack, '\0');
	  buffer->content = (uchar *) obstack_finish (buffer->input_stack);
	} 
	return;
	
      default:
      string_plain_char:
	obstack_1grow (buffer->input_stack, c);
	break;
      }
  }
}

/* buffer_get
 *   if the language is case insensitive,
 *   build a lower case version of the buffer
 */
void
buffer_get (struct buffer * buffer)
{
  if (buffer->stream)
    /* The input method is streams.  Use the obstack to get char
     * by char */
    buffer_stream_get_line (buffer);
  else
    buffer_string_get_line (buffer);
      
  /* One more line read */
  buffer->line++;
  
  /* If the eol char is preceded by a \f, then just forget the eol,
   * so that there won't be a blank line at the top of the next page */
  if ((buffer->len >= 2) 
      && buffer->content [buffer->len - 2] == '\f') {
    buffer->content [--(buffer->len)] = '\0';
  }

  if (buffer->lower_case)
    {
      size_t i;
      if (buffer->allocsize <= buffer->len)
	buffer->allocsize = buffer->len + 1;

      buffer->value = 
	REALLOC (buffer->value, uchar, buffer->allocsize);

      for (i = 0 ; i <= buffer->len ; i++)
	buffer->value[i] = tolower(buffer->content[i]);
    } else {
      buffer->value = buffer->content;
    }
  
  buffer->curr = 0;
}
