#include <strstream.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include "cccc_htm.h"
#include "cccc_utl.h"
#include "cccc.h"

#ifndef USE_SHORT_FILENAMES
#define USE_SHORT_FILENAMES 0
#endif


char curdirbuf[1024];
char updirbuf[1024];

static void Make_Filename_Absolute(CCCC_String& filename)
{
  char fnbuf[1024];
  strcpy(fnbuf,filename);

  // this function is intended to do its best to convert a filename
  // received by the program into as absolute a form as possible
  // so that URLs based on the absolute form will still be valid if the
  // document which contains them moves

  // also there seem to be some filename patterns which Netscape at least
  // is not happy with (i.e. starting with ".")

  // first look for a DOS style filename (with a ":" anywhere)
  // if this is found no changes are needed
  if(strchr(fnbuf,':'))
  {
    // a DOS style path with a device name
    // could be relative after the device
    // DOS users are welcome to add code to handle this
  }
  else if (fnbuf[0]=='/')
  {
    // an absolute Unix style file name,
    // no work required
  }
  else if ( strncmp(fnbuf,"..",2)==0 )
  {
    // filename starting with ..
    // skip to after the first directory separator (DOS or Unix)
    filename=updirbuf;
    filename=filename+dirsep;
    // throw away the part up to the first separator
    char* rest_of_name=&(fnbuf[3]);
    filename=filename+rest_of_name;
  }
  else if(
    (strncmp(fnbuf,"./",2) == 0) ||
    (strncmp(fnbuf,".\\",2) == 0) 
    )
  {
    // filename starting with .
    // skip to after the first directory separator (DOS or Unix)
    filename=curdirbuf;
    filename=filename+dirsep;
    // throw away the part up to the first separator
    char* rest_of_name=&(fnbuf[2]);
    filename=filename+rest_of_name;
  }
  else
  {
    // an unqualified name, hopefully relative to the current directory
    filename=curdirbuf;
    filename=filename+dirsep;
    filename=filename+fnbuf;
  }
}

static const CCCC_String& module_output_filename(
  CCCC_String outdir, CCCC_Module *mod_ptr, int i) 
{
      char mfilename[1000];


#if USE_SHORT_FILENAMES
      sprintf(mfilename,"%s%scccc_%03d.htm",(char*)outdir,dirsep,i);
#else
      sprintf(mfilename,"%s%s%s.htm",
              (char*)outdir,dirsep,mod_ptr->name(nlSIMPLE));
#endif
      static CCCC_String filename;
      filename=mfilename;

#if 0
      // making filenames absolute seemed like a good idea at the time
      // as it would allow the top level document to be moved freely and
      // still refer to the detail reports for each module where they lie
      // unfortunately absolute filenames are no good if (as is likely)
      // cccc runs in a Unix environment but the reports are then browsed
      // using a DOS browser over a NFS or Samba file system
      Make_Filename_Absolute(filename);
#endif
      return filename;
}

CCCC_Html_Stream::CCCC_Html_Stream(
  CCCC_Project& project, const char* outfile, 
  int report_mask, const char* libdir, const char* outdir) 
  : prj(project), fstr(outfile), libdir(libdir), outdir(outdir) 
{ 
  // capture the current working directory for the run and its parent 
  // (this will be used to help provide links which are independent
  // of the location of the generated file)
  getcwd(curdirbuf,1024);
  chdir("..");
  getcwd(updirbuf,1024);
  chdir(curdirbuf);

  if(fstr.good() == 0)
  {
    cerr << "failed to open " << outfile << " for output" << endl;
    exit(1);
  }

  fstr << "<HTML><HEAD><TITLE>" << endl
       << "CCCC Report" << endl;

  if(strcmp(prj.name(nlSIMPLE),"")!=0) 
  {
    fstr << " for project "
	 << prj.name(nlSIMPLE)
	 << endl;
  }

  time(&time_generated);
  fstr << " generated " << ctime(&time_generated) << endl
       << "</TITLE></HEAD>" << endl;
  fstr << "<BODY>" << endl;

  if(report_mask & rtCONTENTS)
  {
    Table_Of_Contents(report_mask);
  }

  if(report_mask & rtSUMMARY)
  {
    Project_Summary();
  }

  if(report_mask & rtOODESIGN)
  {
    OO_Design();
  }

  if(report_mask & rtSTRUCT1)
  {
    Structural_Summary();
  }
    
  if(report_mask & rtSTRUCT2)
  {
    Structural_Detail();
  }

  if(report_mask & rtPROC1)
  {
    Procedural_Summary();
  }

  if(report_mask & rtPROC2)
  {
    Procedural_Detail();
  }
  
  if(report_mask & rtSEPARATE_MODULES)
  {
    Separate_Modules();
  }

  if(report_mask & rtREJECTED)
  {
    Rejected_Extents();
  }

  if(report_mask & rtCCCC)
  {
    Put_Section_Heading("About CCCC","infocccc",1);
    Include_Library_File("cccc_inf.dat");
  }

  fstr << "</BODY></HTML>" << endl; 
}

void CCCC_Html_Stream::Table_Of_Contents(int report_mask)
{
  // record the number of report parts in the table, and the 
  // stream put pointer
  // if we find that we have only generated a single part, we supress
  // the TOC by seeking to the saved stream offset
  int number_of_report_parts=0;
  int saved_stream_offset=fstr.tellp();

  fstr << "<TABLE BORDER WIDTH=100%>" << endl
       << "<TR><TH COLSPAN=2>" << endl
       << "CCCC Software Metrics Report";
  if( strlen(prj.name(nlSIMPLE)) > 0 )
  {
    fstr << " on project " << prj.name(nlSIMPLE); 
  }
  fstr << endl 
       << "<BR> generated " << ctime(&time_generated) << endl

       << "</TR>" << endl;

  if(report_mask & rtSUMMARY)
  {
    Put_Section_TOC_Entry(
      "Project Summary","projsum",
      "Summary table of high level measures summed over all files processed "
      "in the current run.");
    number_of_report_parts++;
  }

  if(report_mask & rtOODESIGN)
  {
    Put_Section_TOC_Entry(
      "Object Oriented Design","oodesign",
      "Table of four of the 6 metrics proposed by Chidamber and Kemerer "
      "in their various papers on "
      "'a metrics suite for object oriented design'.");
    number_of_report_parts++;
  }

  if(report_mask & rtSTRUCT1)
  {
    Put_Section_TOC_Entry(
      "Structural Metrics Summary", "structsum",
      "Structural metrics based on the relationships of each module with "
      "others.  Includes fan-out (i.e. number of other modules the current "
      "module uses), fan-in (number of other modules which use the current "
      "module), and the Information Flow measure suggested by Henry and "
      "Kafura, which combines these to give a measure of coupling for the "
      "module.");
    number_of_report_parts++;
  }
 
  if(report_mask & rtSTRUCT2)
  {
    Put_Section_TOC_Entry(
      "Structural Metrics Detail", "structdet",
      "The names of the modules included as clients and suppliers in the "
      "counts for the Structural Metrics Summary.");
    number_of_report_parts++;
  }

  if(report_mask & rtPROC1)
  {
    Put_Section_TOC_Entry(
      "Procedural Metrics Summary", "procsum",
      "Table of procedural measures (i.e. lines of code, lines of comment, "
      "McCabe's cyclomatic complexity summed over each module.");
    number_of_report_parts++;
  }

  if(report_mask & rtPROC2)
  {
    Put_Section_TOC_Entry(
      "Procedural Metrics Detail", "procdet",
      "The same procedural metrics as in the procedural metrics summary, "
      "reported for individual functions, grouped by module.");
    number_of_report_parts++;
  }
 
  if(report_mask & rtREJECTED)
  {
    Put_Section_TOC_Entry(
      "Rejected Extents", "rejected",
      "Extents of submitted source files which the analyser was unable to parse."
      );
    number_of_report_parts++;
  }

  if(report_mask & rtCCCC)
  {
    Put_Section_TOC_Entry(
      "About CCCC", "infocccc",
      "A description of the CCCC program.");
    number_of_report_parts++;
  }

  fstr << "</TR></TABLE>" << endl;
  if(number_of_report_parts<2)
  {
    fstr.seekp(saved_stream_offset);
  }
}

void CCCC_Html_Stream::Put_Section_Heading(
  const CCCC_String& heading_title, 
  const CCCC_String& heading_tag, 
  int heading_level)
{
  fstr << "<H" << heading_level << ">" 
       << "<A NAME=\"" << heading_tag << "\">" 
       << heading_title 
       << "</A></H" << heading_level
       << ">" << endl;
} 

void  CCCC_Html_Stream::Project_Summary() {
  Put_Section_Heading("Project Summary","projsum",1);

  fstr << "This table shows measures over the project as a whole." << endl;

  fstr << "<UL>" << endl;
  Metric_Description("NOM","Number of modules",
		     "Number of non-trivial modules identified by the "
		     "analyser.  Non-trivial modules include all classes, "
		     "and any other module for which member functions are "
		     "identified.");
  Metric_Description("LOC","Lines of Code",
		     "Number of non-blank, non-comment lines of source code "
		     "counted by the analyser.");
  Metric_Description("COM","Lines of Comments",
		     "Number of lines of comment identified by the analyser");
  Metric_Description("MVG","McCabe's Cyclomatic Complexity",
		     "A measure of the decision complexity of the functions "
		     "which make up the program."
		     "The strict definition of this measure is that it is "
		     "the number of linearly independent routes through "
		     "a directed acyclic graph which maps the flow of control "
		     "of a subprogram.  The analyser counts this by recording "
		     "the number of distinct decision outcomes contained "
		     "within each function, which yields a good approximation "
		     "to the formally defined version of the measure.");
  Metric_Description("L_C","Lines of code per line of comment",
		     "Indicates density of comments with respect to textual "
		     "size of program");
  Metric_Description("M_C","Cyclomatic Complexity per line of comment",
		     "Indicates density of comments with respect to logical "
		     "complexity of program");
  Metric_Description("IF4","Information Flow measure",
		     "Measure of information flow between modules suggested "
		     "by Henry and Kafura. The analyser makes an approximate "
		     "count of this by counting inter-module couplings "
		     "identified in the module interfaces.");

  fstr << "</UL>" << endl
       << "Two variants on the information flow measure IF4 are also "
       << "presented, one (IF4v) calculated using only relationships in the "
       << "visible part of the module interface, and the other (IF4c) "
       << "calculated using only those relationships which imply that changes "
       << "to the client must be recompiled of the supplier's definition "
       <<  "changes." 
       << endl << endl;
    
  // calculate the counts on which all displayed data will be based
  int nom=prj.get_count("NOM");  // number of modules 
  int loc=prj.get_count("LOC");  // lines of code
  int mvg=prj.get_count("MVG");  // McCabes cyclomatic complexity
  int com=prj.get_count("COM");  // lines of comment
  int if4=prj.get_count("IF4");    // intermodule complexity (all couplings)
  int if4v=prj.get_count("IF4v");  // intermodule complexity (visible only)
  int if4c=prj.get_count("IF4c");  // intermodule complexity (concrete only)
  int rej=prj.rejected_extent_table.get_count("LOC");
  
  fstr << "<TABLE BORDER WIDTH=100%>" << endl
       << "<TR>" << endl;
  Put_Header_Cell("Metric",70);
  Put_Header_Cell("Tag",10);
  Put_Header_Cell("Overall",10);
  Put_Header_Cell("Per Module",10);
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Number of modules");
  Put_Label_Cell("NOM");
  Put_Metric_Cell(nom);
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Lines of Code",700);
  Put_Label_Cell("LOC",120);
  Put_Metric_Cell(loc,"LOCp");
  Put_Metric_Cell(loc,nom,"LOCm");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("McCabe's Cyclomatic Number");
  Put_Label_Cell("MVG");
  Put_Metric_Cell(mvg,"MVGp");
  Put_Metric_Cell(mvg,nom,"MVGm");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Lines of Comment");
  Put_Label_Cell("COM");
  Put_Metric_Cell(com);
  Put_Metric_Cell(com,nom);
  fstr << "</TR>" << endl;
  
  fstr << "<TR>" << endl;
  Put_Label_Cell("LOC/COM");
  Put_Label_Cell("L_C");
  Put_Metric_Cell(loc,com,"L_C");
  Put_Label_Cell("");
  fstr << "</TR>" << endl;
  
  fstr << "<TR>" << endl;
  Put_Label_Cell("MVG/COM");
  Put_Label_Cell("M_C");
  Put_Metric_Cell(mvg,com,"M_C");
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (inclusive)");
  Put_Label_Cell("IF4");
  Put_Metric_Cell(if4);
  Put_Metric_Cell(if4,nom,"IF4");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (visible)");
  Put_Label_Cell("IF4v");
  Put_Metric_Cell(if4v);
  Put_Metric_Cell(if4v,nom,"IF4v");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (concrete)");
  Put_Label_Cell("IF4c");
  Put_Metric_Cell(if4c);
  Put_Metric_Cell(if4c,nom,"IF4c");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Lines of Code rejected by parser");
  Put_Label_Cell("REJ");
  Put_Metric_Cell(rej,"REJ");
  Put_Metric_Cell(0);
  fstr << "</TR>" << endl;

  fstr << "</TABLE>" << endl;
}

void CCCC_Html_Stream::OO_Design() {
  Put_Section_Heading("Object Oriented Design","oodesign",1);

  fstr << "<UL>" << endl;
  Metric_Description("WMC","Weighted methods per class",
		     "The sum of a weighting function over the functions of "
		     "the module.  Two different weighting functions are "
		     "applied: WMC1 uses the nominal weight of 1 for each "
		     "function, and hence measures the number of functions, "
		     "WMCv uses a weighting function which is 1 for functions "
		     "accessible to other modules, 0 for private functions.");
  Metric_Description("DIT","Depth of inheritance tree",
		     "The length of the longest path of inheritance ending at "
		     "the current module.  The deeper the inheritance tree "
		     "for a module, the harder it may be to predict its "
		     "behaviour.  On the other hand, increasing depth gives "
		     "the potential of greater reuse by the current module "
		     "of behaviour defined for ancestor classes.");
  Metric_Description("NOC","Number of children",
		     "The number of modules which inherit directly from the "
		     "current module.  Moderate values of this measure "
		     "indicate scope for reuse, however high values may "
		     "indicate an inappropriate abstraction in the design.");
  Metric_Description("CBO","Coupling between objects",
		     "The number of other modules which are coupled to the "
		     "current module either as a client or a supplier. "
		     "Excessive coupling indicates weakness of module "
		     "encapsulation and may inhibit reuse.");
  fstr << "</UL>" << endl << endl;

  fstr << "The label cell for each row in this table provides a link to "
       << "the module summary table in the detailed report for the "
       << "module in question" << endl;


  fstr << "<TABLE BORDER WIDTH=100%>" << endl
       << "<TR>" << endl;
  Put_Header_Cell("Module Name",50);
  Put_Header_Cell("WMC1",10);
  Put_Header_Cell("WMCv",10);
  Put_Header_Cell("DIT",10);
  Put_Header_Cell("NOC",10);
  Put_Header_Cell("CBO",10);

  fstr << "</TR>" << endl;

  int i;
  for(i=0; i<prj.module_table.records();i++)
  {
    CCCC_Module* mod_ptr=(CCCC_Module*)(prj.module_table.record_ptr(i));

    if( mod_ptr->is_trivial() == 0)
    {
      char *metric_tags[5]={"WMC1","WMCv","DIT","NOC","CBO"};
      fstr << "<TR>" << endl;

      CCCC_String href=module_output_filename(outdir,mod_ptr,i);
      href = href+"#summary";

      Put_Label_Cell(mod_ptr->name(nlSIMPLE),0,"",href);
      
      int j;
      for(j=0; j<5; j++)
      {
	CCCC_Metric metric_value(
	  mod_ptr->get_count(metric_tags[j]),
	  metric_tags[j]
	  );
	Put_Metric_Cell(metric_value);
      }
      fstr << "</TR>" << endl;

    }
  }

  fstr << "</TABLE>" << endl;
  
}

void CCCC_Html_Stream::Procedural_Summary() {
  Put_Section_Heading("Procedural Metrics Summary","procsum",1);

  fstr << "For descriptions of each of these metrics see the information "
       << "preceding the project summary table."
       << endl << endl;

  fstr << "The label cell for each row in this table provides a link to "
       << "the functions table in the detailed report for the "
       << "module in question" << endl;

  fstr << "<TABLE BORDER WIDTH=100%>" << endl
       << "<TR>" << endl;
  Put_Header_Cell("Module Name");
  Put_Header_Cell("LOC",8);
  Put_Header_Cell("MVG",8);
  Put_Header_Cell("COM",8);
  Put_Header_Cell("L_C",8);
  Put_Header_Cell("M_C",8);

  fstr << "</TR>" << endl;

  int i;
  for(i=0; i<prj.module_table.records();i++)
  {
    CCCC_Module* mod_ptr=(CCCC_Module*)(prj.module_table.record_ptr(i));

    if( mod_ptr->is_trivial() == 0)
    {
      fstr << "<TR>" << endl;
      CCCC_String href=module_output_filename(outdir,mod_ptr,i);
      href = href+"#procdet";

      Put_Label_Cell(mod_ptr->name(nlSIMPLE),0,"",href);
      int loc=mod_ptr->get_count("LOC");
      int mvg=mod_ptr->get_count("MVG");
      int com=mod_ptr->get_count("COM");
      CCCC_Metric mloc(loc,"LOCm");
      CCCC_Metric mmvg(mvg,"MVGm");
      CCCC_Metric ml_c(loc,com,"L_C");
      CCCC_Metric mm_c(mvg,com,"M_C");

      Put_Metric_Cell(mloc);
      Put_Metric_Cell(mmvg);
      Put_Metric_Cell(com);
      Put_Metric_Cell(ml_c);
      Put_Metric_Cell(mm_c);

      fstr << "</TR>" << endl;

    }
  }

  fstr << "</TABLE>" << endl;
  
}

void CCCC_Html_Stream::Structural_Summary() 
{
  Put_Section_Heading("Structural Metrics Summary","structsum",1);

  fstr << "<UL>" << endl;
  Metric_Description("FI","Fan-in",
		     "The number of other modules which pass information "
		     "into the current module.");
  Metric_Description("FO","Fan-out",
		     "The number of other modules into which the current "
		     "module passes information");
  Metric_Description("IF4","Information Flow measure",
		     "A composite measure of structural complexity, "
		     "calculated as the square of the product of the fan-in "
		     "and fan-out of a single module.  Proposed by Henry and "
		     "Kafura.");
  fstr << "</UL>" << endl;

  fstr << "Note that the fan-in and fan-out are calculated by examining the "
       << "interface of each module.  As noted above, three variants of each "
       << "each of these measures are presented: a count restricted to the "
       << "part of the interface which is externally visible, a count which "
       << "only includes relationships which imply the client module needs "
       << "to be recompiled if the supplier's implementation changes, and an "
       << "inclusive count" << endl << endl;


  fstr << "The label cell for each row in this table provides a link to "
       << "the relationships table in the detailed report for the "
       << "module in question" << endl << endl;

  fstr << "<TABLE BORDER WIDTH=100%>" << endl;
  fstr  << "<TR>" << endl
	<< "<TH BGCOLOR=AQUA ROWSPAN=2>Module Name</TH>" << endl
	<< "<TH BGCOLOR=AQUA COLSPAN=3>Fan-out</TH>" << endl
	<< "<TH BGCOLOR=AQUA COLSPAN=3>Fan-in</TH>" << endl
	<< "<TH BGCOLOR=AQUA COLSPAN=3>IF4</TH>" << endl
	<< "</TR>" << endl;
  Put_Header_Cell("vis",7);
  Put_Header_Cell("con",7);
  Put_Header_Cell("inc",7);
  Put_Header_Cell("vis",7);
  Put_Header_Cell("con",7);
  Put_Header_Cell("incl",7);
  Put_Header_Cell("vis",7);
  Put_Header_Cell("con",7);
  Put_Header_Cell("inc",7);

  fstr << "</TR>" << endl;

  int i=0;
  while(i<prj.module_table.records())
  {
    CCCC_Module* module_ptr=prj.module_table.record_ptr(i);
    if(!module_ptr->is_trivial())
    {
      fstr << "<TR>" << endl;

      int fov=module_ptr->get_count("FOv");
      int foc=module_ptr->get_count("FOc");
      int fo=module_ptr->get_count("FO");

      int fiv=module_ptr->get_count("FIv");
      int fic=module_ptr->get_count("FIc");
      int fi=module_ptr->get_count("FI");

      int if4v=module_ptr->get_count("IF4v");
      int if4c=module_ptr->get_count("IF4c");
      int if4=module_ptr->get_count("IF4");

      // the last two arguments here turn on links to enable jumping between
      // the summary and detail cells for the same module
      CCCC_String href=module_output_filename(outdir,module_ptr,i);
      href = href+"#structdet";      
      Put_Label_Cell(module_ptr->name(nlSIMPLE), 0, "",href);
      Put_Metric_Cell(CCCC_Metric(fov,"FOv"));
      Put_Metric_Cell(CCCC_Metric(foc,"FOc"));
      Put_Metric_Cell(CCCC_Metric(fo,"FO"));
      Put_Metric_Cell(CCCC_Metric(fiv,"FIv"));
      Put_Metric_Cell(CCCC_Metric(fic,"FIc"));
      Put_Metric_Cell(CCCC_Metric(fi,"FI"));
      Put_Metric_Cell(CCCC_Metric(if4v,"IF4v"));
      Put_Metric_Cell(CCCC_Metric(if4c,"IF4c"));
      Put_Metric_Cell(CCCC_Metric(if4,"IF4"));

      fstr << "</TR>" << endl;
    }
    i++;
  }  
  fstr << "</TABLE>" << endl;
  
}

void CCCC_Html_Stream::Put_Structural_Details_Cell(
  CCCC_Module *mod, CCCC_Project *prj, int mask, UserelNameLevel nl)
{
  CCCC_UseRelationship **peers;
  int relnum=mod->get_relationships(prj, mask, peers);

  fstr << "<TD WIDTH=50%>" << endl;
  int j;
  for(j=0; j<relnum; j++)
  {
    CCCC_UseRelationship *ur_ptr=peers[j];
    fstr << ur_ptr->name(nl) << " ";
    AugmentedBool vis=ur_ptr->is_visible();
    AugmentedBool con=ur_ptr->is_concrete();

    if( (vis != abTRUE) && (con != abTRUE) )
    {
      fstr << "[CV] ";
    }
    else if(vis != abTRUE)
    {
      fstr << "[V] ";
    }
    else if(con != abTRUE)
    {
      fstr << "[C] ";
    }
    fstr << "<BR>" << endl;  
    Put_Extent_List(*ur_ptr);
    fstr << "<BR>" << endl;
  }
  // put a non-breaking space in to avoid the unpleasantness which
  // goes with completely empty cells
  fstr << "&nbsp;" << endl;

  fstr << "</TD>" << endl;

}
void CCCC_Html_Stream::Structural_Detail() 
{
  Put_Section_Heading("Structural Metrics Detail","structdet",1);
  fstr << "<TABLE BORDER WIDTH=100%>" << endl;
  fstr  << "<TR>" << endl;
  Put_Header_Cell("Module Name",20);
  Put_Header_Cell("Clients",40);
  Put_Header_Cell("Suppliers",40);
  fstr << "</TR>" << endl;

  int i=0;
  while(i<prj.module_table.records())
  {
    CCCC_Module* module_ptr=prj.module_table.record_ptr(i);
    if(!module_ptr->is_trivial())
    {
      fstr << "<TR>" << endl;
      Put_Label_Cell(module_ptr->name(nlSIMPLE), 0, "structdet","structsum");
      Structural_Detail(module_ptr);
      fstr << "</TR>" << endl;
    }
    i++;
  }  
  fstr << "</TABLE>" << endl;
  
}

void CCCC_Html_Stream::Procedural_Detail() {
  Put_Section_Heading("Procedural Metrics Detail","procdet",1);

  fstr << "<TABLE BORDER WIDTH=100%>" << endl;

  int i;
  for(i=0; i<prj.module_table.records();i++)
  {
    CCCC_Module* mod_ptr=prj.module_table.record_ptr(i);

    if( 
      (strcmp(mod_ptr->name(nlMODULE_TYPE),"builtin") != 0) &&
      (strcmp(mod_ptr->name(nlMODULE_TYPE),"enum") != 0) &&
      (strcmp(mod_ptr->name(nlMODULE_TYPE),"union") != 0) &&
      (mod_ptr->member_table.records()>0) 
      )
    {
      fstr  << "<TR>" << endl;
      Put_Label_Cell(mod_ptr->name(nlSIMPLE),50,"procdet","procsum",mod_ptr);
      Put_Header_Cell("LOC",10);
      Put_Header_Cell("MVG",10);
      Put_Header_Cell("COM",10);
      Put_Header_Cell("L_C",10);
      Put_Header_Cell("M_C",10);

      fstr << "</TR>" << endl;
      Procedural_Detail(mod_ptr);
    }
  }  
  fstr << "</TABLE>" << endl;
}

void CCCC_Html_Stream::Rejected_Extents() 
{
  Put_Section_Heading("Rejected Extents","rejected",1);
  fstr << "<TABLE BORDER WIDTH=100%>" << endl;
  fstr << "<TR>" << endl;
  Put_Header_Cell("Location",25);
  Put_Header_Cell("Text",45);
  Put_Header_Cell("LOC",10);
  Put_Header_Cell("COM",10);
  Put_Header_Cell("MVG",10);
  fstr << "</TR>" << endl;

  if(prj.rejected_extent_table.records() == 0)
  {
    fstr << "<TR><TD COLSPAN=5>"
	 << "No extents were rejected in this run"
	 << "</TD></TR>" << endl;
  }
  else
  {
    int i;
    for(i=0; i<prj.rejected_extent_table.records();i++)
    {
      CCCC_Extent *extent_ptr=prj.rejected_extent_table.record_ptr(i);
      fstr << "<TR>";
      Put_Extent_Cell(*extent_ptr,0);
      Put_Label_Cell(extent_ptr->name(nlDESCRIPTION));
      Put_Metric_Cell(extent_ptr->get_count("LOC"),"");
      Put_Metric_Cell(extent_ptr->get_count("COM"),"");
      Put_Metric_Cell(extent_ptr->get_count("MVG"),"");
      fstr << "</TR>" << endl;
    }
  }
  fstr << "</TABLE>" << endl;
  
}


void CCCC_Html_Stream::Include_Library_File(char* fname) {
  int lines_appended=0;

  MAKE_STRSTREAM(fname_stream);
  fname_stream << libdir << dirsep << fname << ends;
  ifstream istr(fname_stream.str());
  RELEASE_STRSTREAM(fname_stream);

  while(istr.good())
  {
    char buffer[1024];
    istr.getline(buffer,1024);
    fstr << buffer << endl;
    lines_appended++;
  }

  cerr << lines_appended 
       << " lines appended from library file " << fname 
       << endl;
}

void CCCC_Html_Stream::Put_Section_TOC_Entry(
  const CCCC_String& section_name, const CCCC_String& section_href,
  const CCCC_String& section_description)
{
  fstr << "<TR>" << endl
       << "<TH><H4><A HREF=\"#" << section_href << "\">"
       << section_name << "</A></H4></TH>" << endl 
       << "<TD>" << endl
       << section_description << endl
       << "</TR>" << endl;
}

void CCCC_Html_Stream::Put_Header_Cell(const CCCC_String& label, int width)
{
  fstr << "<TH BGCOLOR=\"AQUA\"";
  if(width>0)
  {
    fstr << " WIDTH=" << width << "%" ;
  }
  fstr << ">" ;
  if(strlen(label)>0)
  {
    *this << label;
  }
  else
  {
    // put a non-breaking space in to avoid the strange
    // bevelling associated with empty cells
    fstr << "&nbsp;";

  }
  fstr << "</TH>";
}

void CCCC_Html_Stream::Put_Label_Cell(
  const CCCC_String& label, int width,
  const CCCC_String& ref_name, const CCCC_String& ref_href, 
  CCCC_Record *rec_ptr)
{
  fstr << "<TD";
  if(width>0)
  {
    fstr << " WIDTH=" << width <<"%";
  }
  fstr << ">" ;

  if(strlen(ref_name) > 0)
  {
    // we need to insert an HTML "<A NAME=...> tag for the current cell
    // this enables other locations to jump in
    fstr << "<A NAME=\"" << ref_name << "\"></A>" << endl;
  }

  if(strlen(ref_href) > 0)
  {
    // we need to insert an HTML <A HREF=...> tag for the current cell
    // this enables this cell to be a link to jump out
    fstr << "<A HREF=\"" << ref_href << "\">" << endl;
    // this anchor will need to be closed after the label has been displayed
  }

  if(strlen(label)>0)
  {
    *this << label ;
  }
  else
  {
    // put a non-breaking space in to avoid the strange
    // bevelling associated with empty cells
    fstr << "&nbsp;";
  }

  if(strlen(ref_href) > 0)
  {
    // closing the anchor we opened above
    fstr << "</A>" << endl;
  }

  if(rec_ptr != 0)
  {
    fstr << "<BR>" << endl;
    Put_Extent_List(*rec_ptr);
  }

  fstr << "</TD>";
}
  

void CCCC_Html_Stream::Put_Metric_Cell(
  int count, const CCCC_String& tag, int width)
{
  CCCC_Metric m(count, tag);
  Put_Metric_Cell(m, width);
}

void CCCC_Html_Stream::Put_Metric_Cell(
  int num, int denom, const CCCC_String& tag, int width)
{
  CCCC_Metric m(num,denom, tag);
  Put_Metric_Cell(m, width);
}

void  CCCC_Html_Stream::Put_Metric_Cell(const CCCC_Metric& metric, int width)
{
  fstr << "<TD ALIGN=RIGHT";

  if(width>0)
  {
    fstr << " WIDTH=" << width << "%" ;
  }

  switch(metric.emphasis_level())
  {
  case elMEDIUM:
    fstr << " BGCOLOR=\"YELLOW\"";
    break;
  case elHIGH:
    fstr << " BGCOLOR=\"RED\"";
    break;
  default:
    // no background colour
    break;
  }
  fstr << ">";

  *this << metric;
  fstr << "</TD>";
}

void CCCC_Html_Stream::Put_Extent_URL(const CCCC_Extent& extent)
{
  CCCC_String filename=extent.name(nlFILENAME);
  Make_Filename_Absolute(filename);

  fstr << "<CODE><A HREF=\"" << filename << "\">";
  fstr << extent.name(nlFILENAME) << ":" ;
  fstr << extent.name(nlLINENUMBER) << "</A></CODE>" << endl;
}

void CCCC_Html_Stream::Put_Extent_Cell(const CCCC_Extent& extent, int width) {
  fstr << "<TD";
  if(width>0)
  {
    fstr << " WIDTH=" << width << "%>";
  }
  else
  {
    fstr << ">";
  }
  Put_Extent_URL(extent);
  fstr << "</TD>" << endl;
}

void CCCC_Html_Stream::Put_Extent_List(CCCC_Record& record) 
{
  int i;
  for(i=0; i<record.extent_table.records(); i++)
  {
    CCCC_Extent *ext_ptr=(CCCC_Extent*) ( record.extent_table.record_ptr(i) );
    Put_Extent_URL(*ext_ptr);
  }
  fstr << "<BR>" << endl;
}

// the next two methods define the two basic output operations through which
// all of the higher level output operations are composed
CCCC_Html_Stream& operator <<(CCCC_Html_Stream& os, const CCCC_String& stg) 
{
  // initialise a character pointer to the start of the string's buffer
  char *cptr=stg;
  while(*cptr!='\000') {
    char c=*cptr;
	
    // the purpose of this is to filter out the characters which
    // must be escaped in HTML
    switch(c) {
    case '>': os.fstr << "&gt;" ; break;
    case '<': os.fstr << "&lt;" ; break;
    case '&': os.fstr << "&amp;"; break;
      // commas and parentheses do not need to be escaped, but
      // we want to allow line breaking just inside
      // parameter lists and after commas
      // we insert a non-breaking space to guarantee a small indent
      // on the new line, and one before the right parenthesis for
      // symetry
    case ',': os.fstr << ", &nbsp;" ; break;
    case '(': os.fstr << "( &nbsp;" ; break;
    case ')': os.fstr << "&nbsp;)" ; break;
    default : os.fstr << c;
    }
    cptr++;
  }
  return os;
}

CCCC_Html_Stream& operator <<(CCCC_Html_Stream& os, const CCCC_Metric& mtc) 
{
  char *emphasis_prefix[]={"","<EM>","<STRONG>"};
  char *emphasis_suffix[]={"","</EM>","</STRONG>"};

  // by writing to the underlying ostream object, we avoid the escape
  // functionality
  os.fstr << emphasis_prefix[mtc.emphasis_level()]
	  << mtc.value_string() 
	  << emphasis_suffix[mtc.emphasis_level()];
  return os;
}

CCCC_Html_Stream::CCCC_Html_Stream(
      CCCC_Project& project, CCCC_Module& module, 
      const char* fname)
  : prj(project), fstr(fname) 
{ 
  if(fstr.good() == 0)
  {
    cerr << "failed to open " << fname << " for output" << endl;
    exit(1);
  }
    
  CCCC_String modname=module.name(nlSIMPLE);

  cerr << "Generating report on " << modname << " to " << fname << endl;

  fstr << "<HTML><HEAD><TITLE>" << endl
       << "CCCC Detailed Report on module " << modname << endl;

  time(&time_generated);
  fstr << " generated " << ctime(&time_generated) << endl
       << "</TITLE>" << endl
       << "</HEAD>" << endl
       << "<BODY>" << endl;

  CCCC_String report_title="Detailed report on module ";
  report_title = report_title + modname;
  Put_Section_Heading(report_title,"summary",1);

  fstr << "generated " << ctime(&time_generated) << endl;

  Module_Summary(&module);

  Put_Section_Heading("Relationships","structdet",2);
  fstr << "<TABLE BORDER WIDTH=100%>" << endl;
  fstr << "<TR><TH WIDTH=50%>Clients</TH><TH WIDTH=50%>Suppliers</TH></TR>" 
       << endl
       << "<TR>" << endl;
  Structural_Detail(&module);
  fstr << "</TR></TABLE>" << endl;
   
  Put_Section_Heading("Functions","proc",2);
  fstr  << "<TABLE BORDER WIDTH=100%><TR>" << endl;
  Put_Label_Cell("Function prototype",50);
  Put_Header_Cell("LOC",10);
  Put_Header_Cell("MVG",10);
  Put_Header_Cell("COM",10);
  Put_Header_Cell("L_C",10);
  Put_Header_Cell("M_C",10);
  Procedural_Detail(&module);
  fstr << "</TR></TABLE>" << endl;

}

void CCCC_Html_Stream::Separate_Modules()
{
  // this function generates a separate HTML report for each non-trivial
  // module in the database

  // make a directory to contain the separate files 
  // (it may exist already)
  mode_t dir_create_mode=0755;
  mkdir((const char*) outdir,dir_create_mode);

  int i;
  for(i=0; i<prj.module_table.records();i++)
  {
    CCCC_Module* mod_ptr=prj.module_table.record_ptr(i);

   if(!mod_ptr->is_trivial())
   {
      CCCC_String mfilename=module_output_filename(outdir, mod_ptr, i);
      CCCC_Html_Stream(prj,*mod_ptr,mfilename);
    }
  }
}

void CCCC_Html_Stream::Procedural_Detail(CCCC_Module *module_ptr)
{
  // this function generates the contents of the procedural detail table
  // relating to a single module
  
  // the output needs to be enveloped in a pair of <TABLE></TABLE> tags
  // these have not been put within the function because it is designed
  // to be used in two contexts:
  // 1. within the Separate_Modules function, wrapped directly in the table
  //    tags
  // 2. within the Procedural_Detail function, where the table tags are 
  //    around the output of many calls to this function

  if(module_ptr->member_table.records()==0)
  {
    fstr << "<TR><TD COLSPAN=6>"
	 << "No prodedural members have been identified for this module"
	 << "</TD></TR>" << endl;
  }
  else
  {
    int j;
    for(j=0; j<module_ptr->member_table.records();j++)
    {
      CCCC_Member* mem_ptr=
	(CCCC_Member*)(module_ptr->member_table.record_ptr(j));


      fstr << "<TR>" << endl;
      Put_Label_Cell(mem_ptr->name(nlLOCAL),0,"","",mem_ptr);
      int loc=mem_ptr->get_count("LOC");
      int mvg=mem_ptr->get_count("MVG");
      int com=mem_ptr->get_count("COM");
      CCCC_Metric mloc(loc,"LOCf");
      CCCC_Metric mmvg(mvg,"MVGf");
      CCCC_Metric ml_c(loc,com,"L_C");
      CCCC_Metric mm_c(mvg,com,"M_C");


      Put_Metric_Cell(mloc);
      Put_Metric_Cell(mmvg);
      Put_Metric_Cell(com);
      Put_Metric_Cell(ml_c);
      Put_Metric_Cell(mm_c);
      fstr << "</TR>" << endl;

    }
  }
  fstr << "<TR><TD HEIGHT=12 COLSPAN=6></TD></TR>" << endl;
}

void CCCC_Html_Stream::Metric_Description(
  const CCCC_String& abbreviation, 
  const CCCC_String& name, 
  const CCCC_String& description)
{
  // this is intended to be called in the context of an unnumbered list
  fstr << "<LI>" << abbreviation << " = " << name << "<BR>" << endl
       << description << endl;
}


void CCCC_Html_Stream::Structural_Detail(CCCC_Module *module_ptr) 
{
  
      CCCC_UseRelationship **fi_peers=NULL, **fo_peers=NULL;
      int client_mask=rmeCLIENT|rmeHIDDEN_OR_VISIBLE|rmeABSTRACT_OR_CONCRETE;
      Put_Structural_Details_Cell(module_ptr, &prj, client_mask, nlCLIENT);

      int supplier_mask=
	rmeSUPPLIER|rmeHIDDEN_OR_VISIBLE|rmeABSTRACT_OR_CONCRETE;
      Put_Structural_Details_Cell(module_ptr, &prj, supplier_mask, nlSUPPLIER);

}

void CCCC_Html_Stream::Module_Summary(CCCC_Module *module_ptr) 
{
  // calculate the counts on which all displayed data will be based
  int nof=module_ptr->member_table.records(); // Number of functions
  int loc=module_ptr->get_count("LOC");  // lines of code
  int mvg=module_ptr->get_count("MVG");  // McCabes cyclomatic complexity
  int com=module_ptr->get_count("COM");  // lines of comment

  // the variants of IF4 measure information flow and couplings
  int if4=module_ptr->get_count("IF4");   // (all couplings)
  int if4v=module_ptr->get_count("IF4v"); // (visible only)
  int if4c=module_ptr->get_count("IF4c"); // (concrete only)

  int wmc1=module_ptr->get_count("WMC1"); // Weighted methods/class (unity)
  int wmcv=module_ptr->get_count("WMCv"); // Weighted methods/class (visible)
  int dit=module_ptr->get_count("DIT");   // depth of inheritance tree
  int noc=module_ptr->get_count("NOC");   // number of children
  int cbo=module_ptr->get_count("CBO");   // coupling between objects

  fstr << "<TABLE BORDER WIDTH=100%>" << endl
       << "<TR>" << endl;
  Put_Header_Cell("Metric",70);
  Put_Header_Cell("Tag",10);
  Put_Header_Cell("Overall",10);
  Put_Header_Cell("Per Function",10);
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Lines of Code");
  Put_Label_Cell("LOC");
  Put_Metric_Cell(loc,"LOCm");
  Put_Metric_Cell(loc,nof,"LOCg");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("McCabe's Cyclomatic Number");
  Put_Label_Cell("MVG");
  Put_Metric_Cell(mvg,"MVGm");
  Put_Metric_Cell(mvg,nof,"MVGf");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Lines of Comment");
  Put_Label_Cell("COM");
  Put_Metric_Cell(com);
  Put_Metric_Cell(com,nof);
  fstr << "</TR>" << endl;
  
  fstr << "<TR>" << endl;
  Put_Label_Cell("LOC/COM");
  Put_Label_Cell("L_C");
  Put_Metric_Cell(loc,com,"L_C");
  Put_Label_Cell("");
  fstr << "</TR>" << endl;
  
  fstr << "<TR>" << endl;
  Put_Label_Cell("MVG/COM");
  Put_Label_Cell("M_C");
  Put_Metric_Cell(mvg,com,"M_C");
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (inclusive)");
  Put_Label_Cell("IF4");
  Put_Metric_Cell(if4,1,"IF4");
  Put_Metric_Cell(if4,nof);
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (visible)");
  Put_Label_Cell("IF4v");
  Put_Metric_Cell(if4v,1,"IF4v");
  Put_Metric_Cell(if4v,nof);
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Information Flow measure (concrete)");
  Put_Label_Cell("IF4c");
  Put_Metric_Cell(if4c,1,"IF4c");
  Put_Metric_Cell(if4c,nof);
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Weighted Methods per Class (weighting = unity)");
  Put_Label_Cell("WMC1");
  Put_Metric_Cell(wmc1);
  Put_Label_Cell(""); // wmc1 should be identical to nof
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Weighted Methods per Class (weighting = visible)");
  Put_Label_Cell("WMCv");
  Put_Metric_Cell(wmcv);
  Put_Metric_Cell(wmcv,nof,"WMCv");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Depth of Inheritance Tree");
  Put_Label_Cell("DIT");
  Put_Metric_Cell(dit);
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Number of Children");
  Put_Label_Cell("NOC");
  Put_Metric_Cell(noc);
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "<TR>" << endl;
  Put_Label_Cell("Coupling between objects");
  Put_Label_Cell("CBO");
  Put_Metric_Cell(cbo);
  Put_Label_Cell("");
  fstr << "</TR>" << endl;

  fstr << "</TABLE>" << endl;
}



#ifdef UNIT_TEST
int main()
{
  CCCC_Project *prj_ptr=test_project_ptr();
  CCCC_Html_Stream os(*prj_ptr,"cccc.htm",".");
  return 0;
}
#endif
