/* iconmanager.c - runs stuff when icons are clicked on etc.
   Copyright (C) 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include "coolwidget.h"
#include "icon.h"
#include "mad.h"

extern void destroy_icon (CWidget * w);
CWidget *CDrawPowerIcon (CPowerIcon * p);
int edit_dialog (CPowerIcon * icon, char *heading);
char *substitute_strings (char *text, char *cmdline_options, char *editor_file, char *dnd_type, char *dnd_data_file);

void new_dialog (CWidget * w);
int save_icons (void);
void change_config_file (CWidget * icon);
void edit_change_directory (void);
void raise_lower_all (int r);
void arrange_icons (void);
void close_all (CWidget * w);
void icon_help (void);
void edit_tri_cursor (Window win);


void free_icon_elements (CPowerIcon * icon)
{
    if (icon->title)
	free (icon->title);
    if (icon->xpm_filename)
	free (icon->xpm_filename);
    if (icon->prompt)
	free (icon->prompt);
    if (icon->confirm)
	free (icon->confirm);
    if (icon->drop_script)
	free (icon->drop_script);
    if (icon->double_click_script)
	free (icon->double_click_script);
    if (icon->tool_hint)
	free (icon->tool_hint);
    memset (icon, 0, sizeof (CPowerIcon));
}

void copy_icon_elements (CPowerIcon * icon, CPowerIcon * icon2)
{
    memcpy (icon, icon2, sizeof (CPowerIcon));
    icon->title = strdup (icon->title);
    icon->xpm_filename = strdup (icon->xpm_filename);
    icon->prompt = strdup (icon->prompt);
    icon->confirm = strdup (icon->confirm);
    icon->drop_script = strdup (icon->drop_script);
    icon->double_click_script = strdup (icon->double_click_script);
    icon->tool_hint = strdup (icon->tool_hint);
}

void destroy_power_icon (CWidget * w)
{
    CPowerIcon *icon;
    destroy_icon (w);
    icon = (CPowerIcon *) w->user;
    if (!icon)
	return;
    free_icon_elements (icon);
    free (icon);
    w->user = 0;
}

void CDestroyIconWidget (char *ident)
{
    CWidget *w;
    w = CIdent (ident);
    if (w->droppedmenu) {
	CPullUp (w->droppedmenu);
	CDestroyWidget (w->droppedmenu->ident);
	w->droppedmenu = 0;
    }
    CDestroyWidget (ident);
}

int savefile (const char *filename, const char *data, int len, int permissions);

void icon_system (char *script, char *temp_data_file)
{
    pid_t p;
    p = fork ();
    if (p == -1) {
	CErrorDialog (CRoot, 20, 20, _ (" Execute Script "), _ (" Fork Failed "));
    } else {
	if (!p) {
	    char *temp_script_file;
#ifdef HAVE_MAD
	    temp_script_file = strdup (tempnam ("/tmp", "cicon"));
#else
	    temp_script_file = tempnam ("/tmp", "cicon");
#endif
	    savefile (temp_script_file, script, strlen (script), 0700);
	    CSystem (temp_script_file);
	    remove (temp_script_file);
	    if (*temp_data_file)
		remove (temp_data_file);
	    exit (0);
	}
    }
}

void run_script (char *script, char *arg, char *drop_data, int dnd_len, int dnd_type)
{
    char *f, *temp_data_file = 0;
    char *dndtypes[] =
    {
	"DndNotDnd",
	"DndUnknown",
	"DndRawData",
	"DndFile",
	"DndFiles",
	"DndText",
	"DndDir",
	"DndLink",
	"DndExe",
	"DndURL",
	"DndMIME"
    };
    if (dnd_type >= DndEND)
	dnd_type = DndNotDnd;
    if (dnd_type == DndFile)
	f = drop_data;
    else
	f = "";
    if (dnd_type != -1) {
#ifdef HAVE_MAD
	temp_data_file = strdup (tempnam ("/tmp", "cicon"));
#else
	temp_data_file = tempnam ("/tmp", "cicon");
#endif
	if (dnd_type == DndFiles) {
	    char *p;
	    p = (char *) drop_data;
	    for (;;) {
		if ((!p[0] && !p[1]) || (unsigned long) p - (unsigned long) drop_data >= dnd_len)
		    break;
		if (!*p)
		    *p = '\n';
		p++;
	    }
	}
	savefile (temp_data_file, drop_data, dnd_len, 0600);
    } else {
	temp_data_file = strdup ("");
    }
    script = substitute_strings (script, arg, f, dndtypes[dnd_type + 1], temp_data_file);

    icon_system (script, temp_data_file);

    free (script);
    free (temp_data_file);
}

void execute_icon_double_click (CEvent * cevent, CPowerIcon * icon)
{
    char *a = 0;
    if (!*icon->double_click_script)
	return;
    CSetOverrideRedirect ();
    if (icon->options & ICON_OPTION_CONFIRM_DOUBLE_CLICK) {
	switch (CQueryDialog (CRoot, icon->x, icon->y, _(" Execute "), catstrs (" ", icon->confirm, " ", 0), _(" Ok "), _(" Cancel "), 0)) {
	case -1:
	case 1:
	    CClearOverrideRedirect ();
	    return;
	case 0:
	    break;
	}
    }
    CClearOverrideRedirect ();

    if (icon->options & ICON_OPTION_PROMPT_DOUBLE_CLICK) {
	a = CInputDialog ("iconinput", CRoot, icon->x, icon->y, 300 | INPUT_DIALOG_BROWSE, CLastInput ("iconinput"), _(" Execute icon "), icon->prompt);
	if (!a)
	    return;
    }

    CHourGlass (cevent->window);
    run_script (icon->double_click_script, a ? a : "", "", 0, -1);
    edit_tri_cursor (cevent->window);
    if (a)
	free (a);
}

void execute_icon (CWidget * w)
{
    CEvent cevent;
    memset (&cevent, 0, sizeof (CEvent));
    cevent.window = w->winid;
    execute_icon_double_click (&cevent, (CPowerIcon *) w->user);
}

void execute_icon_drop (CEvent *cevent, CPowerIcon * icon, char *drop_data, int dnd_len, int dnd_type)
{
    char *a = 0;
    if (!*icon->drop_script)
	return;
    CSetOverrideRedirect ();
    if (icon->options & ICON_OPTION_CONFIRM_DROP) {
	switch (CQueryDialog (CRoot, icon->x, icon->y, _(" Execute "), catstrs (" ", icon->confirm, " ", 0), _(" Ok "), _(" Cancel "), 0)) {
	case -1:
	case 1:
	    CClearOverrideRedirect ();
	    return;
	case 0:
	    break;
	}
    }
    CClearOverrideRedirect ();
    if (icon->options & ICON_OPTION_PROMPT_DROP) {
	a = CInputDialog ("iconinput", CRoot, icon->x, icon->y, 300 | INPUT_DIALOG_BROWSE, CLastInput ("iconinput"), _(" Execute icon "), icon->prompt);
	if (!a)
	    return;
    }

    CHourGlass (cevent->window);
    run_script (icon->drop_script, a ? a : "", drop_data, dnd_len, dnd_type);
    edit_tri_cursor (cevent->window);
    if (a)
	free (a);
}

void copy_icon (CWidget * w)
{
    CPowerIcon icon;
    copy_icon_elements (&icon, (CPowerIcon *) w->user);
    icon.x = w->x + 10;
    icon.y = w->y + 10;
    CDrawPowerIcon (&icon);
}

void edit_icon (CWidget *w)
{
    CPowerIcon icon;
    copy_icon_elements (&icon, (CPowerIcon *) w->user);
    CDisable (w->ident);
    if (edit_dialog (&icon, _(" Edit Icon "))) {
	CDestroyIconWidget (w->ident);
	CDrawPowerIcon (&icon);
	save_icons ();
    } else {
	CEnable (w->ident);
	free_icon_elements (&icon);
    }
}

void delete_icon (CWidget * w)
{
    XUngrabPointer (CDisplay, CurrentTime);
    CSetOverrideRedirect ();
    switch (CQueryDialog (CRoot, w->x, w->y, _(" Delete "), _(" Remove icon from desktop : "), _(" Remove "), _(" Cancel "), 0)) {
    case -1:
    case 1:
	CClearOverrideRedirect ();
	return;
    case 0:
	break;
    }
    CClearOverrideRedirect ();
    CDestroyIconWidget (w->ident);
}

void icon_menu (char *ident)
{
    CEvent cwevent;
    XEvent xevent;
    CWidget *w;
    w = CIdent (ident);
    if (!w)
	return;
    if (!w->droppedmenu)
	return;
    CPullDown (w->droppedmenu);
    XGrabPointer (CDisplay, w->droppedmenu->droppedmenu->winid, False,
		  ButtonPressMask | ButtonReleaseMask | ButtonMotionMask,
		  GrabModeSync, GrabModeAsync, CRoot,
		  CGetCursorID (DndEND + 2), CurrentTime);
    for (;;) {
	XAllowEvents (CDisplay, SyncPointer, CurrentTime);
	CNextEvent (&xevent, &cwevent);
	if (xevent.type == ButtonRelease)
	    break;
    }
    XUngrabPointer (CDisplay, CurrentTime);
    w = CIdent (ident);
    if (!w)
	return;
    if (!w->droppedmenu)
	return;
    CPullUp (w->droppedmenu);
}

int icon_callback (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    CPowerIcon *ic;
    char ident[33];
    ic = (CPowerIcon *) w->user;
    ic->x = w->x;
    ic->y = w->y;
    switch (cwevent->type) {
    case ButtonPress:
	if (cwevent->button == Button2) {
	    Window win;
	    win = w->winid;
	    strcpy (ident, cwevent->ident);
	    icon_menu (ident);
	    memset (cwevent, 0, sizeof (CEvent));
	    XLowerWindow (CDisplay, win);
	} else if (cwevent->double_click) {
	    execute_icon_double_click (cwevent, ic);
	}
	break;
    case ClientMessage:{
	    int xs, ys, data_type;
	    unsigned char *data;
	    unsigned long size;
	    data_type = CGetDrop (xevent, &data, &size, &xs, &ys);
	    if (data_type == DndNotDnd)
		return 0;
	    execute_icon_drop (cwevent, ic, (char *) data, size, data_type);
	    if (data)
		free (data);
	    return 1;
	}
	return 0;
    }
    return 0;
}

CWidget *CDrawPowerIcon (CPowerIcon * p)
{
    static int n = 0;
    char *f;
    CWidget *w;
    n++;
    if (p->y + FONT_PIX_PER_LINE > HeightOfScreen (DefaultScreenOfDisplay (CDisplay)))
	p->y = HeightOfScreen (DefaultScreenOfDisplay (CDisplay)) - FONT_PIX_PER_LINE;
    if (p->x + 20 > WidthOfScreen (DefaultScreenOfDisplay (CDisplay)))
	p->x = WidthOfScreen (DefaultScreenOfDisplay (CDisplay)) - 20;
    if (*p->xpm_filename != '/')
	f = catstrs (LIBDIR, "/", p->xpm_filename, 0);
    else
	f = p->xpm_filename;
    w = CDrawIcon (catstrs ("_icon", itoa (n), 0), CRoot, p->x, p->y, f, p->title);
    if (!w)
	return 0;
    if (*p->tool_hint)
/* Toolhint */
	CSetToolHint (catstrs ("_icon", itoa (n), 0), p->tool_hint);
    w->user = CMalloc (sizeof (CPowerIcon));
    memcpy (w->user, p, sizeof (CPowerIcon));
    w->destroy = destroy_power_icon;
    w->callback = icon_callback;
    if (!strcmp (p->double_click_script, "<icon manager>")) {
	w->droppedmenu = CDrawMenuButton (catstrs ("_icon", itoa (n), ".menu", 0), w->winid, CRoot, 0, w->height, AUTO_WIDTH, 1, 9,
					_ (" Dummy "),
					_ ("Help..."), 0, (void (*)(unsigned long)) icon_help, 0,
					_ ("New icon..."), 0, new_dialog, (unsigned long) w,
					_ ("Auto arrange"), 0, arrange_icons, 0,
					_ ("Change directory..."), 0, edit_change_directory, 0,
					_ ("Save icons"), 0, save_icons, 0,
					_ ("Config file..."), 0, change_config_file, (unsigned long) w,
					_ ("Raise icons"), 0, raise_lower_all, 1,
					_ ("Lower icons"), 0, raise_lower_all, 0,
					_ ("Kill Coolicons..."), 0, close_all, (unsigned long) w
	    );
    } else {
	w->droppedmenu = CDrawMenuButton (catstrs ("_icon", itoa (n), ".menu", 0), w->winid, CRoot, 0, w->height, AUTO_WIDTH, 1, 7,
					_ (" Dummy "),
					_ ("Edit icon..."), 0, (void (*)(unsigned long)) edit_icon, (unsigned long) w,
					_ ("Execute icon"), 0, execute_icon, (unsigned long) w,
					_ ("Copy icon"), 0, copy_icon, (unsigned long) w,
					_ ("Delete icon"), 0, delete_icon, (unsigned long) w,
					_ ("New icon..."), 0, new_dialog, (unsigned long) w,
					_ ("Auto arrange"), 0, arrange_icons, 0,
					_ ("Save icons"), 0, save_icons, 0
	    );
    }
    return w;
}

void strnlcpy (unsigned char *s, unsigned char *t)
{
    for (;;) {
	*s = (*t == '\n') ? 129 : *t;
	if (!*t)
	    break;
	s++;
	t++;
    }
}

void memlncpy (unsigned char *s, unsigned char *t, int l)
{
    while (l--) {
	*s = (*t == 129) ? '\n' : *t;
	s++;
	t++;
    }
}

char *gather_icon_options_for_config (void)
{
    char *t, *p;
    int i = 0, size = 0;
    while (last_widget > i++)
	if (CIndex (i) != NULL)
	    if ((CIndex (i))->kind == C_ICON_WIDGET && (CIndex (i))->user) {
		CPowerIcon *icon;
		icon = (CPowerIcon *) (CIndex (i))->user;
		size += strlen (icon->title);
		size += strlen (icon->xpm_filename);
		size += strlen (icon->prompt);
		size += strlen (icon->confirm);
		size += strlen (icon->drop_script);
		size += strlen (icon->double_click_script);
		size += strlen (icon->tool_hint);
		size += 128;
	    }
    p = t = CMalloc (size);
    i = 0;
    while (last_widget > i++)
	if (CIndex (i) != NULL)
	    if ((CIndex (i))->kind == C_ICON_WIDGET && (CIndex (i))->user) {
		CPowerIcon *icon;
		icon = (CPowerIcon *) (CIndex (i))->user;
		strcpy (p, "title = ");
		p += strlen (p);
		strnlcpy ((unsigned char *) p, (unsigned char *) icon->title);
		strcat (p, "\nxpm_filename = ");
		strcat (p, icon->xpm_filename);
		strcat (p, "\nx = ");
		strcat (p, itoa ((CIndex (i))->x));
		strcat (p, "\ny = ");
		strcat (p, itoa ((CIndex (i))->y));
		strcat (p, "\nprompt = ");
		p += strlen (p);
		strnlcpy ((unsigned char *) p, (unsigned char *) icon->prompt);
		strcat (p, "\nconfirm = ");
		p += strlen (p);
		strnlcpy ((unsigned char *) p, (unsigned char *) icon->confirm);
		strcat (p, "\noptions = ");
		strcat (p, itoa (icon->options));
		strcat (p, "\ndrop_script = ");
		p += strlen (p);
		strnlcpy ((unsigned char *) p, (unsigned char *) icon->drop_script);
		strcat (p, "\ndouble_click_script = ");
		p += strlen (p);
		strnlcpy ((unsigned char *) p, (unsigned char *) icon->double_click_script);
		strcat (p, "\ntool_hint = ");
		strcat (p, icon->tool_hint);
		strcat (p, "\n\n");
		p += strlen (p);
	    }
    *p = '\0';
    return t;
}

static char *advance_equate (char *p, char *g, int *l)
{
    char *q;
    while (strchr ("\n\t ", *p))
	p++;
    if (strncmp (p, g, strlen (g)))
	return 0;
    p += strlen (g);
    q = strchr (p, '\n');
    if (!q)
	return 0;
    *l = (unsigned long) q - (unsigned long) p;
    return p;
}

/* returns 1 on error */
static int read_string (char **p, char *g, char **s)
{
    char *q;
    int l;
    q = advance_equate (*p, g, &l);
    if (!q)
	return 1;
    *s = CMalloc (l + 1);
    memlncpy ((unsigned char *) *s, (unsigned char *) q, l);
    (*s)[l] = '\0';
    *p = q + l + 1;
    return 0;
}

/* returns 1 on error */
static int read_int (char **p, char *g, int *i)
{
    char *q;
    int l;
    q = advance_equate (*p, g, &l);
    if (!q)
	return 1;
    *i = atoi (q);
    *p = q + l + 1;
    return 0;
}

/* returns 1 on error */
int draw_icons_from_config (char *t)
{
    char *p;
    int n = 0;
    p = t;
    for (;;) {
	CPowerIcon icon;
	memset (&icon, 0, sizeof (CPowerIcon));
	n++;
	p = strstr (p, "title = ");
	if (!p)
	    break;
	if (read_string (&p, "title = ", &icon.title))
	    goto icon_error;
	if (read_string (&p, "xpm_filename = ", &icon.xpm_filename))
	    goto icon_error;
	if (read_int (&p, "x = ", &icon.x))
	    goto icon_error;
	if (read_int (&p, "y = ", &icon.y))
	    goto icon_error;
	if (read_string (&p, "prompt = ", &icon.prompt))
	    goto icon_error;
	if (read_string (&p, "confirm = ", &icon.confirm))
	    goto icon_error;
	if (read_int (&p, "options = ", &icon.options))
	    goto icon_error;
	if (read_string (&p, "drop_script = ", &icon.drop_script))
	    goto icon_error;
	if (read_string (&p, "double_click_script = ", &icon.double_click_script))
	    goto icon_error;
	if (read_string (&p, "tool_hint = ", &icon.tool_hint))
	    goto icon_error;
	if (*p != '\n')
	    break;
	CDrawPowerIcon (&icon);
	p++;
    }
    return 0;
  icon_error:
    CErrorDialog (CRoot, 20, 20, _ (" Reading Config "), _ (" Error interpreting icon %d from config, \n not bothering to read further "), n);
    return 1;
}

