/* menu.c - proper 3D menus
   Copyright (C) 1996, 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
 */


#include <config.h>
#include <stdio.h>
#include <my_string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include "lkeysym.h"

#include "stringtools.h"
#include "app_glob.c"
#include "coolwidget.h"
#include "coollocal.h"

#include "mad.h"

#define MENU_USER_MODE 1
#define MENU_DRAGGING_MODE 2

int menu_mode = MENU_USER_MODE;
CWidget *current_pulled_button = 0;

int eh_menu (CWidget * w, XEvent * xevent, CEvent * cwevent);
void CMenuSelectionDialog (CWidget * button);

static void render_menu_button (CWidget * wdt);

void destroy_menu (CWidget * w)
{
    int i;
    if (!w)
	return;
    if (!w->menu)
	return;
    for (i = 0; i < w->numlines; i++)
	if (w->menu[i].text)
	    free (w->menu[i].text);
    free (w->menu);
}

/* outermost bevel */
#define BEVEL_MAIN	2
/* next-outermost bevel */
#define BEVEL_IN 	4
#define BEVEL_OUT	5
/* between items, and between items and next-outermost bevel */
#define SPACING		4
#define S		SPACING
/* between window border and items */
#define O		(BEVEL_OUT + SPACING)
/* size of bar item */
#define BAR_HEIGHT	8
#define B		BAR_HEIGHT
/* between items rectangle and text */
#define RELIEF		(TEXT_RELIEF + 1)
/* total height of an item */
#define H		(FONT_PIX_PER_LINE + RELIEF * 2)

/* 
   Gets y1 and y2 of the entry of a menu (i.e. a menu item). x1, y1,
   x2, y2 would describe the rectangle that encloses that entry,
   relative to the top left corner of the pulled-down menu window.
 */
void get_menu_item_extents (int j, struct menu_item m[], int *y1, int *y2)
{
    int i, n_items = 0, n_bars = 0;
    if (j)
	for (i = 0; i < j; i++)
	    if (m[i].text[2])
		n_items++;
	    else
		n_bars++;
    *y1 = O + n_items * (H + S) + n_bars * (B + S) + (m[j].text[2] ? 0 : 2);
    *y2 = *y1 + (m[j].text[2] ? H : (B - 4));
}

/*
   Returns 0-(n-1) where n is the number of entries in that menu. Returns
   which menu item the pointer is in if (x,y) is the pointer pos. Returns
   -1 if out of bounds or between menu items.
 */
int whereis_pointer (int x, int y, int w, int n, struct menu_item m[])
{
    int i, y1, y2;
    if (x >= O && x < w - O)
	for (i = 0; i < n; i++) {
	    if (!m[i].text[2])
		continue;
	    get_menu_item_extents (i, m, &y1, &y2);
	    if (y >= y1) {
		if (y < y2)
		    return i;
	    } else
		break;
	}
    return -1;
}

static int find_menu_hotkey (struct menu_item m[], int this, int num)
{
    unsigned char used_keys[256];
    int n = 0, j;
    for (j = 0; j < num; j++)
	if (m[j].hot_key && j != this)
	    used_keys[n++] = my_lower_case (m[j].hot_key);
    return find_letter_at_word_start ((unsigned char *) m[this].text + 1, used_keys, n);
}

void menu_draw (Window win, int w, int h, struct menu_item m[], int n, int light)
{
    int i, y1, y2;
    static int last_light = 0, last_n = 0;
    static Window last_win = 0;
    render_bevel (win, 0, 0, w - 1, h - 1, BEVEL_MAIN, 0);
    render_bevel (win, BEVEL_IN, BEVEL_IN, w - 1 - BEVEL_IN, h - 1 - BEVEL_IN, BEVEL_OUT - BEVEL_IN, 1);

    if (last_win == win && last_n != n) {
	XClearWindow (CDisplay, win);
    } else if (last_light >= 0 && last_light < n) {
	get_menu_item_extents (last_light, m, &y1, &y2);
	CSetColor (COLOR_FLAT);
	CRectangle (win, O - 1, y1 - 1, w - O * 2 + 2, y2 - y1 + 2);
    }
    last_win = win;
    last_n = n;
    for (i = 0; i < n; i++) {
	get_menu_item_extents (i, m, &y1, &y2);
	if (i == light && m[i].text[2]) {
	    CSetColor (color_widget (11));
	    CRectangle (win, O + 1, y1 + 1, w - O * 2 - 2, y2 - y1 - 2);
	    render_bevel (win, O - 1, y1 - 1, w - O, y2, 2, 0);
	} else {
	    if (m[i].text[2])
		render_bevel (win, O, y1, w - O - 1, y2 - 1, 1, 1);
	    else
		render_bevel (win, O + 6, y1, w - O - 1 - 6, y2 - 1, 2, 0);
	}
	if (m[i].text[2]) {

	    char *u;
	    u = strrchr (m[i].text, '\t');
	    if (u)
		*u = 0;
	    CSetColor (COLOR_BLACK);
	    if (m[i].hot_key == '~')
		m[i].hot_key = find_menu_hotkey (m, i, n);
	    drawstring_xy_hotkey (win, RELIEF + O,
				  RELIEF + y1, m[i].text, m[i].hot_key);
	    if (u) {
		drawstring_xy (win, RELIEF + O + (w - (O + RELIEF) * 2 - CStringWidth (u + 1)),
			       RELIEF + y1, u + 1);
		*u = '\t';
	    }
	}
    }
    last_light = light;
}

void render_menu (CWidget * w)
{
    int n, y1, y2, i;
    if (!w)
	return;
    n = w->numlines;
    get_menu_item_extents (n - 1, w->menu, &y1, &y2);
    w->height = y2 + O;
    w->width = 0;
    for (i = 0; i < n; i++) {
	int t;
	t = CStringWidth (w->menu[i].text);
	t += CStringWidth ("W");
	if (w->width < t)
	    w->width = t;
    }
    w->width += (O + RELIEF) * 2;
    XResizeWindow (CDisplay, w->winid, w->width, w->height);

    if (w->y + w->height > HeightOfScreen (DefaultScreenOfDisplay (CDisplay))) {
	int y;
	y = HeightOfScreen (DefaultScreenOfDisplay (CDisplay)) - w->height;
	if (y < 0)
	    y = 0;
	XMoveWindow (CDisplay, w->winid, w->x, y);
    }
    w->droppedmenu->current = w->current;
    menu_draw (w->winid, w->width, w->height, w->menu, w->numlines, w->current);
}

/* gets a windows position relative to the origin if some ancestor windows,
   window can be of a widget or not */
void CGetWindowPosition (Window win, Window ancestor, int *x_return, int *y_return)
{
    CWidget *w = (CWidget *) 1;
    int x = 0, y = 0;
    Window root, parent, *children;
    unsigned int nchildren, width, height, bd, depth;
    *x_return = *y_return = 0;
    if (win == ancestor)
	return;
    for (;;) {
	if (w)
	    w = CWidgetOfWindow (win);
	if (w)
	    if (w->parentid == CRoot)
		w = 0;
	if (w) {
	    parent = w->parentid;
	    x = w->x;
	    y = w->y;
	} else {
	    if (!XQueryTree (CDisplay, win, &root, &parent, &children, &nchildren))
		return;
	    if (children)
		XFree ((char *) children);
	    XGetGeometry (CDisplay, win, &root, &x, &y, &width, &height, &bd, &depth);
	}
	*x_return += x;
	*y_return += y;
	if (parent == ancestor || parent == CRoot)
	    break;
	win = parent;
    }
}

void focus_stack_remove_window (Window w);

void pull_up (CWidget * button)
{
    if (!button)
	return;
    if (button->kind != C_MENU_BUTTON_WIDGET)
	return;
    if (button->droppedmenu) {
	current_pulled_button = 0;
	CDestroyWidget (button->droppedmenu->ident);
	button->droppedmenu = 0;
    }
    focus_stack_remove_window (button->winid);
}

static CWidget *last_menu = 0;

void CSetLastMenu (CWidget *button)
{
    last_menu = button;
}

CWidget *CGetLastMenu (void)
{
    return last_menu;
}

void menu_hand_cursor (Window win);

static CWidget *pull_down (CWidget * button) /* must create a new widget */
{
    CWidget *menu;
    CWidget *sib;
    int width, height, n;
    int x, y;

    if(button->droppedmenu)
	return 0;

    CSetLastMenu (button);

    sib = button;

    while((sib = CNextFocus (sib)) != button) /* pull up any other sibling menus */
	pull_up (sib);

    n = button->numlines;

    CGetWindowPosition (button->winid, CRoot, &x, &y);

    height = 2;	/* don't know what these are yet */
    width = 2;

    x += button->firstcolumn;

    menu = CSetupWidget (catstrs(button->ident, ".pull", 0), CRoot, x,
	    y + button->height, width, height, C_MENU_WIDGET, INPUT_KEY, COLOR_FLAT, 0);

    menu_hand_cursor (menu->winid);

/* no destroy 'cos gets ->menu gets destroyed by other menu-button-widget */
    menu->numlines = n;
    menu->menu = button->menu;
    menu->eh = eh_menu;
    menu->droppedmenu = button;
    button->droppedmenu = menu;
    current_pulled_button = button;
    return menu;
}

void CPullDown (CWidget * button)
{
    pull_down (button);
}

void CPullUp (CWidget * button)
{
    pull_up (button);
}

static void grab_ungrab (Window win)
{
    XUngrabPointer (CDisplay, CurrentTime);
}

static void pull_down_ungrab (CWidget * w)
{
    CWidget *d;
    d = pull_down (w);
    if (!d)
	return;
    CFocus (w);
    if (menu_mode == MENU_DRAGGING_MODE)
	grab_ungrab (w->winid);
}

int execute_item (CWidget * menu, int item)
{
    int r = 0;
    char ident[33];
    strcpy (ident, menu->ident);
    menu->droppedmenu->current = item;
    XUnmapWindow (CDisplay, menu->winid);
    if (item >= 0 && item < menu->numlines)
	if (menu->menu[item].call_back) {
	    menu->droppedmenu->current = item;
	    (*(menu->menu[item].call_back)) (menu->menu[item].data);
	    r = 1;
	}
    menu = CIdent (ident);	/* menu may not exists anymore */
    if (menu)
	pull_up (menu->droppedmenu);
    CFocusLast ();
    return r;
}

static void render_menu_button (CWidget * wdt)
{
    int w = wdt->width, h = wdt->height;
    int x = 0, y = 0;

    Window win = wdt->winid;

    if (wdt->disabled)
	goto disabled;
    if (wdt->options & BUTTON_PRESSED) {
	render_bevel (win, x, y, x + w - 1, y + h - 1, 2, 1);
    } else if (wdt->options & BUTTON_HIGHLIGHT) {
	CSetColor (COLOR_FLAT);
	XDrawRectangle (CDisplay, win, CGC, x + 1, y + 1, w - 3, h - 3);
	render_bevel (win, x, y, x + w - 1, y + h - 1, 1, 0);
    } else {
  disabled:
	CSetColor (COLOR_FLAT);
	XDrawRectangle (CDisplay, win, CGC, x, y, w - 1, h - 1);
	XDrawRectangle (CDisplay, win, CGC, x + 1, y + 1, w - 3, h - 3);
    }

    if (!wdt->label)
	return;
    if (!(*(wdt->label)))
	return;
    CSetColor (COLOR_BLACK);
    drawstring_xy_hotkey (win, x + 2 + BUTTON_RELIEF, y + 2 + BUTTON_RELIEF, wdt->label, wdt->hotkey);
}


int is_focus_change_key (KeySym k);
int is_focus_prev_key (KeySym k, unsigned int state);

int eh_menubutton (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    int c;
    CWidget *f;
    switch (xevent->type) {
    case FocusOut:
	if (w->droppedmenu)	/* hide the menu if the user switches focus to another application */
	    XUnmapWindow (CDisplay, w->droppedmenu->winid);
    case FocusIn:
	render_menu_button (w);
	CExposeWindowArea (w->parentid, 0, w->x - WIDGET_FOCUS_RING, w->y - WIDGET_FOCUS_RING, w->width + WIDGET_FOCUS_RING * 2, w->height + WIDGET_FOCUS_RING * 2);
	break;
    case KeyPress:
	c = CKeySym (xevent);
	if (!w->droppedmenu) {
	    if (c == XK_space || c == XK_Down || c == XK_KP_Down || c == XK_Return || c == XK_KP_Enter) {
		CMenuSelectionDialog (w);
		return 1;
	    }
	}
	if (c == XK_Escape) {
	    pull_up (w);
	    CFocusLast ();
	    return 1;
	}
	if ((c == XK_KP_Up || c == XK_Up) && w->droppedmenu) {
	    if (w->droppedmenu->current == -1)
		w->droppedmenu->current = 0;
	    do {
		w->droppedmenu->current = (w->droppedmenu->current + w->droppedmenu->numlines - 1) % w->droppedmenu->numlines;
	    } while (w->droppedmenu->menu[w->droppedmenu->current].text[2] == 0);
	    render_menu (w->droppedmenu);
	    return 1;
	}
	if ((c == XK_Down || c == XK_KP_Down) && w->droppedmenu) {
	    do {
		w->droppedmenu->current = (w->droppedmenu->current + 1) % w->droppedmenu->numlines;
	    } while (w->droppedmenu->menu[w->droppedmenu->current].text[2] == 0);
	    render_menu (w->droppedmenu);
	    return 1;
	}
	if (is_focus_prev_key (c, xevent->xkey.state)) {
	    f = CPreviousFocus (w);
	    while (f->kind != C_MENU_BUTTON_WIDGET && (unsigned long) f != (unsigned long) w)
		f = CPreviousFocus (f);
	    if (f) {
		CFocus (f);
		if (w->droppedmenu)
		    CMenuSelectionDialog (f);
	    }
	    return 1;
	}
	if (is_focus_change_key (c)) {
	    f = CNextFocus (w);
	    while (f->kind != C_MENU_BUTTON_WIDGET && (unsigned long) f != (unsigned long) w)
		f = CNextFocus (f);
	    if (f) {
		CFocus (f);
		if (w->droppedmenu)
		    CMenuSelectionDialog (f);
	    }
	    return 1;
	}
	if (w->droppedmenu && c) {
	    int i;
	    if (c == XK_Return || c == XK_space) {
		return execute_item (w->droppedmenu, w->droppedmenu->current);
	    } else {
		for (i = 0; i < w->droppedmenu->numlines; i++)
		    if (match_hotkey (c, w->droppedmenu->menu[i].hot_key))
			return execute_item (w->droppedmenu, i);
	    }
	}
	if (c != XK_Down && c != XK_KP_Down)
	    return 0;
    case ButtonPress:
	if (xevent->type == ButtonPress) {
	    w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
	    w->options |= BUTTON_PRESSED;
	}
	render_menu_button (w);
	grab_ungrab (w->winid);	/* stops the pointer from being contained in the window */
	if (!w->droppedmenu)
	    CMenuSelectionDialog (w);
	return 1;
	break;
    case ButtonRelease:
	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
	w->options |= BUTTON_HIGHLIGHT;
	render_menu_button (w);
	return 1;
    case MotionNotify:
	if (!w->droppedmenu && menu_mode == MENU_DRAGGING_MODE)
	    pull_down_ungrab (w);
	return 1;
    case EnterNotify:
	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
	w->options |= BUTTON_HIGHLIGHT;
	render_menu_button (w);
	break;
    case Expose:
	if (xevent->xexpose.count)
	    break;
    case LeaveNotify:
	w->options &= 0xFFFFFFFFUL - BUTTON_PRESSED - BUTTON_HIGHLIGHT;
	render_menu_button (w);
	break;
    }
    return 0;
}



int eh_menu (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    static Window win = 0;
    static int current = -30000;
    switch (xevent->type) {
    case MotionNotify:
	w->current = whereis_pointer (xevent->xmotion.x, xevent->xmotion.y, w->width, w->numlines, w->menu);
	if (w->current == current && w->winid == win)
	    break;
	current = w->current;
	win = w->winid;
	render_menu (w);
	break;
    case ButtonRelease:
	return execute_item (w, whereis_pointer (xevent->xmotion.x, xevent->xmotion.y, w->width, w->numlines, w->menu));
    case ButtonPress:
	w->current = whereis_pointer (xevent->xmotion.x, xevent->xmotion.y, w->width, w->numlines, w->menu);
	render_menu (w);
	break;
    case Expose:
	if (xevent->xexpose.count)
	    break;
    case LeaveNotify:
	current = w->current = w->droppedmenu->current;
	render_menu (w);
	break;
    }
    return 0;
}

CWidget *CDrawMenuButton (const char *ident, Window parent, Window focus_return,
	int x, int y, int width, int height, int num_items, const char *label,
	const char *text, int hot_key, callfn call_back, unsigned long data,...)
{
    va_list ap;
    CWidget *wdt;
    struct menu_item *m;
    int i;
    int w, h;

    if (width == AUTO_WIDTH || height == AUTO_HEIGHT)
	CTextSize (&w, &h, label);
    if (width == AUTO_WIDTH)
	width = w + 4 + BUTTON_RELIEF * 2;
    if (height == AUTO_HEIGHT)
	height = h + 4 + BUTTON_RELIEF * 2;

    wdt = CSetupWidget (ident, parent, x, y,
	    width, height, C_MENU_BUTTON_WIDGET, INPUT_BUTTON | PointerMotionMask, COLOR_FLAT, 1);

    set_hint_pos (x + width + WIDGET_SPACING, y + height + WIDGET_SPACING);

    wdt->label = strdup (label);
    wdt->hotkey = find_hotkey (wdt);
    wdt->options |= WIDGET_TAKES_FOCUS_RING | WIDGET_HOTKEY_ACTIVATES;

    m = CMalloc (num_items * sizeof (struct menu_item));

    m[0].text = strdup (catstrs(" ", text, " ", 0));
    m[0].hot_key = hot_key;
    m[0].call_back = call_back;
    m[0].data = data;

    va_start (ap, data);
    if (num_items > 1)
	for (i = 1; i < num_items; i++) {
	    m[i].text = strdup (catstrs(" ", va_arg (ap, char *), " ", 0));
	    m[i].hot_key = va_arg (ap, char);
	    m[i].call_back = va_arg (ap, callfn);
	    m[i].data = va_arg (ap, unsigned long);
	}
    va_end (ap);

    wdt->destroy = destroy_menu;
    wdt->numlines = num_items;
    wdt->menu = m;
    wdt->eh = eh_menubutton;

    return wdt;
}

void CAddMenuItem (const char *ident, const char *text, int hot_key, callfn call_back, unsigned long data)
{
    struct menu_item *m;
    CWidget *w;

    w = CIdent (ident);

    m = CMalloc ((w->numlines + 1) * sizeof (struct menu_item));
    memcpy (m, w->menu, w->numlines * sizeof (struct menu_item));
    free (w->menu);
    w->menu = m;
    m[w->numlines].text = strdup (catstrs(" ", text, " ", 0));
    m[w->numlines].hot_key = hot_key;
    m[w->numlines].call_back = call_back;
    m[w->numlines].data = data;

    w->numlines ++;

    if (w->droppedmenu != 0) {
	w->droppedmenu->menu = m;
	w->droppedmenu->numlines = w->numlines;
	w->droppedmenu->current = w->current;
	render_menu (w->droppedmenu);
    }
}


static void remove_item (CWidget * w, int i)
{
    if (i >= w->numlines)
	return;
    if (w->menu[i].text)
	free (w->menu[i].text);
    w->numlines--;
    memmove (&w->menu[i], &w->menu[i + 1], (w->numlines - i) * sizeof (struct menu_item));
    if (w->current == i)
	w->current = -1;
    else if (w->current > i)
	w->current--;
    if (w->droppedmenu != 0) {
	w->droppedmenu->numlines = w->numlines;
	w->droppedmenu->current = w->current;
    }
}

void CRemoveMenuItemNumber (const char *ident, int i)
{
    CWidget *w;
    w = CIdent (ident);
    remove_item (w, i);
}

/*
   Starts from the bottom of the menu and searches for the first
   menu item containing text (strstr != NULL), and deletes it.
   (this is untested).
 */
void CRemoveMenuItem (const char *ident, const char *text)
{
    struct menu_item *m;
    CWidget *w;
    int i;

    w = CIdent (ident);

    if (!w->numlines)
	return;

    m = w->menu;

    for (i = w->numlines - 1; i >= 0; i--) {
	if (!text)
	    goto remove;
	if (strstr (m[i].text, text) || !*text) {
	  remove:;
	    remove_item (w, i);
	    render_menu (w->droppedmenu);
	    return;
	}
    }
}

void CMenuSelectionDialog (CWidget * button)
{
    CEvent cwevent;

    if (menu_mode == MENU_DRAGGING_MODE) {
	pull_down_ungrab (button);
	return;
    }
    if (!button)
	return;

    if (!button->droppedmenu)
	pull_down_ungrab (button);
    else
	return;

    menu_mode = MENU_DRAGGING_MODE;

    while (current_pulled_button) {
	CNextEvent (0, &cwevent);
	if (cwevent.kind == C_MENU_BUTTON_WIDGET || cwevent.kind == C_MENU_WIDGET) {
	    continue;
	} else {
	    if ((cwevent.type == ButtonPress || cwevent.type == ButtonRelease) && current_pulled_button) {
		pull_up (current_pulled_button);
		current_pulled_button = 0;
	    }
	}
    }
    CFocusLast ();
    menu_mode = MENU_USER_MODE;
}


